@@ -20,9 +20,9 @@ static ssize_t read_full_n(int fd, void *data, size_t size)
 	return all_ret;
 }
 
-void map_refresh(int fd, int onceonly __attr_unused__, const char **base,
+void map_refresh(int fd, int onceonly ATTR_UNUSED, const char **base,
 		 unsigned long *len, unsigned long newlen,
-		 const char *name, const char *mboxname __attr_unused__)
+		 const char *name, const char *mboxname ATTR_UNUSED)
 {
 	ssize_t ret;
 	void *p;
@@ -45,7 +45,7 @@ void map_refresh(int fd, int onceonly __attr_unused__, const char **base,
 	*len = ret;
 }
 
-void map_free(const char **base, unsigned long *len __attr_unused__)
+void map_free(const char **base, unsigned long *len ATTR_UNUSED)
 {
 	char *x = (char *) *base;
 
